@@ -25,7 +25,7 @@ package Spreadsheet::Read;
 use strict;
 use warnings;
 
-our $VERSION = "0.61";
+our $VERSION = "0.62";
 sub  Version { $VERSION }
 
 use Carp;
@@ -38,21 +38,29 @@ use File::Temp   qw( );
 use Data::Dumper;
 
 my @parsers = (
-    [ csv  => "Text::CSV_XS",              "0.71"  ],
-    [ csv  => "Text::CSV_PP",              "1.17"  ],
-    [ csv  => "Text::CSV",                 "1.17"  ],
-    [ ods  => "Spreadsheet::ReadSXC",      "0.20"  ],
-    [ sxc  => "Spreadsheet::ReadSXC",      "0.20"  ],
-    [ xls  => "Spreadsheet::ParseExcel",   "0.34"  ],
-    [ xlsx => "Spreadsheet::ParseXLSX",    "0.13"  ],
-    [ xlsx => "Spreadsheet::XLSX",         "0.13"  ],
-    [ prl  => "Spreadsheet::Perl",         ""      ],
+    [ csv  => "Text::CSV_XS",				"0.71"		],
+    [ csv  => "Text::CSV_PP",				"1.17"		],
+    [ csv  => "Text::CSV",				"1.17"		],
+    [ ods  => "Spreadsheet::ReadSXC",			"0.20"		],
+    [ sxc  => "Spreadsheet::ReadSXC",			"0.20"		],
+    [ xls  => "Spreadsheet::ParseExcel",		"0.34"		],
+    [ xlsx => "Spreadsheet::ParseXLSX",			"0.13"		],
+    [ xlsx => "Spreadsheet::XLSX",			"0.13"		],
+    [ xlsx => "Spreadsheet::XLSX::Reader::LibXML",	"v0.38.8"	],
+    [ prl  => "Spreadsheet::Perl",			""		],
 
     # Helper modules
-    [ ios  => "IO::Scalar",                ""      ],
-    [ dmp  => "Data::Peek",                ""      ],
+    [ ios  => "IO::Scalar",				""		],
+    [ dmp  => "Data::Peek",				""		],
     );
-my %can = map { $_->[0] => 0 } @parsers;
+my %can = map {
+    my $preset = $ENV{"SPREADSHEET_READ_\U$_->[0]"};
+    if ($preset) {
+	eval "require $preset";
+	$@ and $preset = "!$preset";
+	}
+    $_->[0] => $preset || "";
+    } @parsers;
 for (@parsers) {
     my ($flag, $mod, $vsn) = @$_;
     $can{$flag} and next;
@@ -126,6 +134,7 @@ sub _parser
 sub parses
 {
     my $type = _parser (shift)	or  return 0;
+    $can{$type} =~ m/^!/ and return 0;
     return $can{$type};
     } # parses
 
@@ -271,6 +280,89 @@ sub _xls_fill
     return _xls_color ($bg);
     } # _xls_fill
 
+sub _xlsx_libxml
+{
+    my $oBook = shift;
+    my @names = @{$oBook->get_worksheet_names};
+    $oBook->{SheetCount} = scalar @names;
+    $oBook->{Worksheet}  = [ ];
+    my %mm;
+    while (my $wks = $oBook->worksheet) {
+	push @{$oBook->{Worksheet}}, $wks;
+	$wks->{Cells}  = [];
+	$wks->{Name}   = shift @names;
+	($wks->{MinRow}, $wks->{MaxRow}) = $wks->row_range;
+	($wks->{MinCol}, $wks->{MaxCol}) = $wks->col_range;
+	foreach my $r ($wks->{MinRow} .. $wks->{MaxRow}) {
+	    foreach my $c ($wks->{MinCol} .. $wks->{MaxCol}) {
+		my $cell = $wks->get_cell ($r, $c);
+		if (defined $cell) {
+		    $cell->{Val}    = $cell->unformatted;
+		    $cell->{Merged} = $cell->is_merged and
+			$mm{$cell->{cell_merge}}{"$r:$c"} = $cell;
+		    $cell->{Type}   = $cell->type;
+		    $cell->{Hidden} = 0;#$cell->is_hidden; NYI
+		    }
+		else {
+		    $cell = {
+			Val    => undef,
+			Type   => "Undef",
+			Merged => 0,
+			Hidden => 0,
+			};
+		    }
+		$wks->{Cells}[$r][$c] = $cell;
+		}
+	    }
+	# Spreadsheet::XLSX::Reader::LibXML returns all other cells in range as undef
+	foreach my $mm (values %mm) {
+	    my @mr = sort keys %$mm;
+	    foreach my $rc (@mr) {
+		defined $mm->{$rc}{Val} or $mm->{$rc}{Val} = "";
+		}
+	    }
+	}
+    *Spreadsheet::XLSX::Reader::LibXML::Cell::Value = sub {
+	my $cell = shift;
+	$cell->value;
+	};
+    *WorksheetInstance::get_merged_areas = sub {
+	my $mm = shift->_merge_map or return;
+	# [ undef,
+	#   [ undef,
+	#     undef,
+	#     'B1:C2',
+	#     'B1:C2'
+	#     ],
+	#   [ undef,
+	#     'A2:A3',
+	#     'B1:C2',
+	#     'B1:C2'
+	#     ],
+	#   [ undef,
+	#     'A2:A3'
+	#     ]
+	#   ]
+	# ->
+	# [ [ 1, 0,	# A2:
+	#     2, 0,	#  A3
+	#     ],
+	#   [ 0, 1,	# B1:
+	#     1, 2,	#  C2
+	#     ]
+	#   ]
+	my %r;
+	for (@$mm) { $_ && $r{$_}++ for @$_ }
+	keys %r or return;
+	my @r;
+	foreach my $ma (keys %r) {
+	    my ($ul, $br) = split m/:/ => $ma or return;
+	    push @r, [ reverse map { $_ - 1 } map { cell2cr ($_) } $br, $ul ];
+	    }
+	return \@r;
+	};
+    } # _xlsx_libxml
+
 sub ReadData
 {
     my $txt = shift	or  return;
@@ -385,16 +477,16 @@ sub ReadData
 	close $in;
 
 	for (@{$data[1]{cell}}) {
-	    defined $_ or $_ = [];
+	    defined or $_ = [];
 	    }
 	return _clipsheets \%opt, [ @data ];
 	}
 
     # From /etc/magic: Microsoft Office Document
     if ($io_txt && _parser ($opt{parser}) !~ m/^xlsx?$/ &&
-		    $txt =~ m/^(\376\067\0\043
+		    $txt =~ m{^(\376\067\0\043
 			       |\320\317\021\340\241\261\032\341
-			       |\333\245-\0\0\0)/x) {
+			       |\333\245-\0\0\0)}x) {
 	$can{xls} or croak "Spreadsheet::ParseExcel not installed";
 	my $tmpfile;
 	if ($can{ios}) { # Do not use a temp file if IO::Scalar is available
@@ -414,20 +506,22 @@ sub ReadData
     if ($opt{parser} ? ($_parser = _parser ($opt{parser})) =~ m/^xlsx?$/
 		     : ($io_fil && $txt =~ m/\.(xlsx?)$/i && ($_parser = $1))) {
 	my $parse_type = $_parser =~ m/x$/i ? "XLSX" : "XLS";
-	$can{lc $parse_type} or croak "Parser for $parse_type is not installed";
+	my $parser = $can{lc $parse_type} or
+	    croak "Parser for $parse_type is not installed";
+	my $xlsx_libxml = $parser =~ m/LibXML$/;
 	$debug and print STDERR "Opening $parse_type \$txt\n";
 	my $oBook = eval {
 	    $io_ref
 	      ? $parse_type eq "XLSX"
-		? $can{xlsx} =~ m/::XLSX/
-		? Spreadsheet::XLSX->new ($io_ref)
-		: Spreadsheet::ParseXLSX->new (%parser_opts)->parse ($io_ref)
-		: Spreadsheet::ParseExcel->new (%parser_opts)->Parse ($io_ref)
+		? $can{xlsx} =~ m/::XLSX$/
+		? $parser->new ($io_ref)
+		: $parser->new (%parser_opts)->parse ($io_ref)
+		: $parser->new (%parser_opts)->Parse ($io_ref)
 	      : $parse_type eq "XLSX"
-		? $can{xlsx} =~ m/::XLSX/
-		? Spreadsheet::XLSX->new ($txt)
-		: Spreadsheet::ParseXLSX->new (%parser_opts)->parse ($txt)
-		: Spreadsheet::ParseExcel->new (%parser_opts)->Parse ($txt);
+		? $can{xlsx} =~ m/::XLSX$/
+		? $parser->new ($txt)
+		: $parser->new (%parser_opts)->parse ($txt)
+		: $parser->new (%parser_opts)->Parse ($txt);
 	    };
 	unless ($oBook) {
 	    # cleanup will fail on folders with spaces.
@@ -435,14 +529,13 @@ sub ReadData
 	    croak "$parse_type parser cannot parse data: $msg";
 	    }
 	$debug > 8 and _dump (oBook => $oBook);
+
+	$xlsx_libxml and _xlsx_libxml ($oBook);
+
 	my @data = ( {
 	    type	=> lc $parse_type,
 	    parser	=> $can{lc $parse_type},
-	    version	=> $parse_type eq "XLSX"
-			 ? $can{xlsx} =~ m/::XLSX/
-			 ? $Spreadsheet::XLSX::VERSION
-			 : $Spreadsheet::ParseXLSX::VERSION
-			 : $Spreadsheet::ParseExcel::VERSION,
+	    version	=> $can{lc $parse_type}->VERSION,
 	    error	=> undef,
 	    sheets	=> $oBook->{SheetCount} || 0,
 	    sheet	=> {},
@@ -456,8 +549,10 @@ sub ReadData
 	    );
 	$oBook->{FormatStr}{$_} = $def_fmt{$_} for keys %def_fmt;
 	my $oFmt = $parse_type eq "XLSX"
-	    ? $can{xlsx} =~ m/::XLSX/
+	    ? $can{xlsx} =~ m/::XLSX$/
 	    ? Spreadsheet::XLSX::Fmt2007->new
+	    : $can{xlsx} =~ m/::XLSX/
+	    ? Spreadsheet::XLSX::Reader::LibXML::FmtDefault->new
 	    : Spreadsheet::ParseExcel::FmtDefault->new
 	    : Spreadsheet::ParseExcel::FmtDefault->new;
 
@@ -473,7 +568,9 @@ sub ReadData
 		merged  => [],
 		);
 	    defined $sheet{label}  or  $sheet{label}  = "-- unlabeled --";
+	    exists $oWkS->{MinRow} and $sheet{minrow} = $oWkS->{MinRow} + 1;
 	    exists $oWkS->{MaxRow} and $sheet{maxrow} = $oWkS->{MaxRow} + 1;
+	    exists $oWkS->{MinCol} and $sheet{mincol} = $oWkS->{MinCol} + 1;
 	    exists $oWkS->{MaxCol} and $sheet{maxcol} = $oWkS->{MaxCol} + 1;
 	    $sheet{merged} = [
 		map  {  $_->[0] }
@@ -500,9 +597,10 @@ sub ReadData
 		foreach my $r ($oWkS->{MinRow} .. $sheet{maxrow}) {
 		    foreach my $c ($oWkS->{MinCol} .. $sheet{maxcol}) {
 			my $oWkC = $oWkS->{Cells}[$r][$c] or next;
-			defined (my $val = $oWkC->{Val})  or next;
+			#defined (my $val = $oWkC->{Val}) or next;
+			my $val = $oWkC->{Val};
 			my $cell = cr2cell ($c + 1, $r + 1);
-			$opt{rc}    and $sheet{cell}[$c + 1][$r + 1] = $val;	# Original
+			$opt{rc} and $sheet{cell}[$c + 1][$r + 1] = $val;	# Original
 
 			my $fmt;
 			my $FmT = $oWkC->{Format};
@@ -521,6 +619,14 @@ sub ReadData
 			    }
 			exists $FmT->{Fill} or $FmT->{Fill} = [ 0 ];
 			exists $FmT->{Font} or $FmT->{Font} = undef;
+			if ($xlsx_libxml and
+			     ref $oWkC eq "Spreadsheet::XLSX::Reader::LibXML::Cell" and
+			     my $style = $oWkC->get_fill) {
+			    if (my $pf = $style->{patternFill}) {
+				#defined $pf->{fgColor}{indexed} and $FmT->{Font}{Color} =   $pf->{fgColor}{indexed};
+				#defined $pf->{bgColor}{indexed} and $FmT->{Fill}        = [ $pf->{bgColor}{indexed} ];
+				}
+			    }
 
 			unless (defined $fmt) {
 			    $fmt = $FmT->{FmtIdx}
@@ -544,9 +650,9 @@ sub ReadData
 			    }
 			defined $fmt and $fmt =~ s/\\//g;
 			$opt{cells} and	# Formatted value
-			    $sheet{$cell} = $FmT && exists $def_fmt{$FmT->{FmtIdx}}
+			    $sheet{$cell} = defined $val ? $FmT && exists $def_fmt{$FmT->{FmtIdx}}
 				? $oFmt->ValFmt ($oWkC, $oBook)
-				: $oWkC->Value;
+				: $oWkC->Value : undef;
 			if ($opt{attr}) {
 			    my $FnT = $FmT->{Font};
 			    my $fmi = $FmT->{FmtIdx}
@@ -558,8 +664,8 @@ sub ReadData
 
 				type    => lc $oWkC->{Type},
 				enc     => $oWkC->{Code},
-				merged  => $oWkC->is_merged || 0,
-				hidden  => $FmT->{Hidden}   || 0,
+				merged  => (defined $oWkC->{Merged} ? $oWkC->{Merged} : $oWkC->is_merged) || 0,
+				hidden  => (defined $oWkC->{Hidden} ? $oWkC->{Hidden} : $FmT->{Hidden})   || 0,
 				locked  => $FmT->{Lock}     || 0,
 				format  => $fmi,
 				halign  => [ undef, qw( left center right
@@ -576,12 +682,13 @@ sub ReadData
 				fgcolor => _xls_color ($FnT->{Color}),
 				bgcolor => _xls_fill  (@{$FmT->{Fill}}),
 				};
+			    #_dump "cell", $sheet{attr}[$c + 1][$r + 1];
 			    }
 			}
 		    }
 		}
 	    for (@{$sheet{cell}}) {
-		defined $_ or $_ = [];
+		defined or $_ = [];
 		}
 	    push @data, { %sheet };
 #	    $data[0]{sheets}++;
@@ -646,7 +753,7 @@ sub ReadData
 	    # r67c7 = [P2L] 2*(1000*r67c5-60)
 	    }
 	for (@{$data[1]{cell}}) {
-	    defined $_ or $_ = [];
+	    defined or $_ = [];
 	    }
 	return _clipsheets \%opt, [ @data ];
 	}
@@ -717,7 +824,7 @@ sub ReadData
 			}
 		    }
 		for (@{$sheet{cell}}) {
-		    defined $_ or $_ = [];
+		    defined or $_ = [];
 		    }
 		$debug and print STDERR "\tSheet $sheet_idx '$sheet{label}' $sheet{maxrow} x $sheet{maxcol}\n";
 		push @data, { %sheet };
@@ -733,6 +840,7 @@ sub ReadData
 
 1;
 
+__END__
 =head1 DESCRIPTION
 
 Spreadsheet::Read tries to transparently read *any* spreadsheet and
@@ -742,7 +850,8 @@ module that does the actual spreadsheet scanning.
 For OpenOffice and/or LibreOffice this module uses L<Spreadsheet::ReadSXC>
 
 For Microsoft Excel this module uses L<Spreadsheet::ParseExcel>,
-L<Spreadsheet::ParseXLSX>, or L<Spreadsheet::XLSX> (discouraged).
+L<Spreadsheet::ParseXLSX>, L<Spreadsheet::XLSX::Reader::LibXML>, or
+L<Spreadsheet::XLSX> (discouraged).
 
 For CSV this module uses L<Text::CSV_XS> or L<Text::CSV_PP>.
 
@@ -838,6 +947,13 @@ C<ods>, C<openoffice>, C<libreoffice>) C<xls> (or C<excel>), and C<xlsx>
 When parsing streams, instead of files, it is highly recommended to pass
 this option.
 
+Spreadsheet::Read supports several underlying parsers per spreadsheet
+type. It will try those from most favored to least favored. When you
+have a good reason to prefer a different parser, you can set that in
+environment variables. The other options then will not be tested for:
+
+ env SPREADSHEET_READ_CSV=Text::CSV_PP ...
+
 =item cells
 
 Control the generation of named cells ("A1" etc). Default is true.
@@ -906,29 +1022,6 @@ parser.
 All other attributes/options will be passed to the underlying parser if
 that parser supports attributes.
 
-=back
-
-=head2 Using CSV
-
-In case of CSV parsing, C<ReadData ()> will use the first line of the file
-to auto-detect the separation character if the first argument is a file and
-both C<sep> and C<quote> are not passed as attributes. Text::CSV_XS (or
-Text::CSV_PP) is able to automatically detect and use C<\r> line endings).
-
-CSV can parse streams too, but be sure to pass C<sep> and/or C<quote> if
-these do not match the default C<,> and C<">.
-
-When an error is found in the CSV, it is automatically reported (to STDERR).
-The structure will store the error in C<< $ss->[0]{error} >> as anonymous
-list returned by C<< $csv->error_diag >>. See Text::CSV_XS for documentation.
-
- my $ss = ReadData ("bad.csv");
- $ss->[0]{error} and say $ss->[0]{error}[1];
-
-=head2 Functions
-
-=over 4
-
 =item my $cell = cr2cell (col, row)
 
 C<cr2cell ()> converts a C<(column, row)> pair (1 based) to the
@@ -1000,6 +1093,23 @@ use argument list, or call it fully qualified.
 
 =back
 
+=head2 Using CSV
+
+In case of CSV parsing, C<ReadData ()> will use the first line of the file
+to auto-detect the separation character if the first argument is a file and
+both C<sep> and C<quote> are not passed as attributes. Text::CSV_XS (or
+Text::CSV_PP) is able to automatically detect and use C<\r> line endings.
+
+CSV can parse streams too, but be sure to pass C<sep> and/or C<quote> if
+these do not match the default C<,> and C<">.
+
+When an error is found in the CSV, it is automatically reported (to STDERR).
+The structure will store the error in C<< $ss->[0]{error} >> as anonymous
+list returned by C<< $csv->error_diag >>. See Text::CSV_XS for documentation.
+
+ my $ss = ReadData ("bad.csv");
+ $ss->[0]{error} and say $ss->[0]{error}[1];
+
 =head2 Cell Attributes
 
 If the constructor was called with C<attr> having a true value, effort
@@ -1280,6 +1390,12 @@ http://metacpan.org/release/Spreadsheet-ParseExcel
 
 http://metacpan.org/release/Spreadsheet-ParseXLSX
 
+=item Spreadsheet::XLSX::Reader::LibXML
+
+http://metacpan.org/release/Spreadsheet-XLSX-Reader-LibXML
+
+Work in progress. Much is relying on undocumented internals.
+
 =item Spreadsheet::XLSX
 
 http://metacpan.org/release/Spreadsheet-XLSX
